import pandas
from enda.models import ModelInterface

try:
    from sklearn.linear_model import LinearRegression
except ImportError:
    raise ImportError("h2o is required is you want to use this enda's H2OModel. Try: pip install scikit-learn>=0.24.1")


class SKLearnLinearRegression(ModelInterface):
    """ The simplest model, serving as an example of an implementation of ModelInterface """

    def __init__(self):
        self.lin_reg_model = None

    def train(self, df: pandas.DataFrame, target_col: str):
        x = df.drop(columns=[target_col])
        y = df[target_col]
        self.lin_reg_model = LinearRegression().fit(x, y)

    def predict(self, df: pandas.DataFrame, target_col: str):
        a = self.lin_reg_model.predict(df)  # numpy array
        s = pandas.Series(a, name=target_col, index=df.index)  # pandas series with correct name and index
        return s.to_frame()



import pandas as pd
from pandas.api.types import is_string_dtype, is_datetime64_dtype
from dateutil.relativedelta import relativedelta
from enda.timezone_utils import TimezoneUtils


class Contracts:
    """
    A class to help handle contracts data.

    contracts : a dataframe with a contract or sub-contract on each row.
        A contract row has fixed-values between two dates. In some systems, some values of a contract can change over
        time, for instance the "subscribed power" of an electricity consumption contract.
        Different ERPs can handle it in different ways, for instance with "sub-contracts" inside a contract
        or with contract "amendments".
        Here each row is a period between two dates where all the values of the contract are fixed.

        Columns can be like in example_a.csv :
        [customer_id,
         contract_id,
         date_start,
         date_end_exclusive,  # contract ends at 00h00 that day, so that day is excluded from the contract period.
         sub_contract_end_reason,
         subscribed_power_kva,
         smart_metered,
         profile,
         customer_type,
         specific_price,
         estimated_annual_consumption_kwh,
         tension]

    Columns [date_start, date_end_exclusive] are required (names can differ), and with dtype=datetime64 (tz-naive).
    Other columns describe contract characteristics.


    """

    @classmethod
    def read_contracts_from_file(cls,
                                 file_path,
                                 date_start_col="date_start",
                                 date_end_exclusive_col="date_end_exclusive",
                                 date_format="%Y-%m-%d"):
        """
        Reads contracts from a file. This will convert start and end date columns into dtype=datetime64 (tz-naive)
        :param file_path: where the source file is located.
        :param date_start_col: the name of your contract date start column.
        :param date_end_exclusive_col: the name of your contract date end column, end date is exclusive.
        :param date_format: the date format for pandas.to_datetime.
        :return: a pandas.DataFrame with a contract on each row.
        """

        df = pd.read_csv(file_path)
        for c in [date_start_col, date_end_exclusive_col]:
            if is_string_dtype(df[c]):
                df[c] = pd.to_datetime(df[c], format=date_format)
        cls.check_contracts_dates(df, date_start_col, date_end_exclusive_col)
        return df

    @staticmethod
    def check_contracts_dates(df, date_start_col, date_end_exclusive_col):
        """
        Checks that the two columns are present, with dtype=datetime64 (tz-naive)
        Checks that date_start is always present.
        A date_end_exclusive==NaT means that the contract has no limited duration.
        Checks that date_start < date_end_exclusive when date_end_exclusive is set

        :param df: the pandas.DataFrame containing the contracts
        :param date_start_col: the name of your contract date start column.
        :param date_end_exclusive_col: the name of your contract date end column, end date is exclusive.
        """

        # check required columns and their types
        for c in [date_start_col, date_end_exclusive_col]:
            if c not in df.columns:
                raise ValueError("Required column not found : {}".format(c))
            # data at a 'daily' precision should not have TZ information
            if not is_datetime64_dtype(df[c]):
                raise ValueError("Expected tz-naive datetime dtype for column, but found dtype: {}"
                                 .format(c, df[c].dtype))
        # check NaN values for date_start
        if df[date_start_col].isnull().any():
            rows_with_nan_date_start = df[df[date_start_col].isnull()]
            raise ValueError("There are NaN values in these rows:\n{}".format(rows_with_nan_date_start))

        contracts_with_end = df.dropna(subset=[date_end_exclusive_col])
        not_ok = contracts_with_end[date_start_col] >= contracts_with_end[date_end_exclusive_col]
        if not_ok.sum() >= 1:
            raise ValueError("Some contracts end before they start:\n{}".format(contracts_with_end[not_ok]))

    @staticmethod
    def __contract_to_events(contracts, date_start_col, date_end_exclusive_col):
        # check that no column is named "event_type" or "event_date"
        for c in ["event_type", "event_date", "date"]:
            if c in contracts.columns:
                raise ValueError("contracts has a column named {}, but this name is reserved in this"
                                 "function; rename your column.".format(c))

        columns_to_keep = [c for c in contracts.columns if c not in [date_start_col, date_end_exclusive_col]]
        events_columns = ["event_type", "event_date"] + columns_to_keep

        # compute "contract start" and "contract end" events
        start_contract_events = contracts.copy(deep=True)  # all contracts must have a start date
        start_contract_events["event_type"] = "start"
        start_contract_events["event_date"] = start_contract_events[date_start_col]
        start_contract_events = start_contract_events[events_columns]

        # for "contract end" events, only keep contracts with an end date (NaT = contract is not over)
        end_contract_events = contracts[contracts[date_end_exclusive_col].notna()].copy(deep=True)
        end_contract_events["event_type"] = "end"
        end_contract_events["event_date"] = end_contract_events[date_end_exclusive_col]
        end_contract_events = end_contract_events[events_columns]

        # concat all events together and sort them chronologically
        all_events = pd.concat([start_contract_events, end_contract_events])
        all_events.sort_values(by=["event_date", "event_type"], inplace=True)

        return all_events

    @classmethod
    def compute_portfolio_by_day(
            cls,
            contracts,
            columns_to_sum,
            date_start_col="date_start",
            date_end_exclusive_col="date_end_exclusive",
            max_date_exclusive=None,
    ):
        """
        Given a list of contracts_with_group ,

        Returns the "portfolio" by day, which is, for each day from the first start of a contract
        to the last end of a contract, the quantities in "columns_to_sum" over time.
        See unittests or enda's guides for examples.

        If you want to compute the quantities for a group of customers, filter contracts before using this function.

        :param contracts: the dataframe containing the list of contracts to consider.
            Each contract has at least these columns :
                ["date_start", "date_end_exclusive", and all columns_to_sum]
                each column in columns_to_sum must be of a summable dtype
        :param columns_to_sum: the columns on which to compute a running-sum over time
        :param date_start_col:
        :param date_end_exclusive_col:
        :param max_date_exclusive: restricts the output to strictly before this date.
                                   Useful if you have end_dates far in the future.

        :return: a 'portfolio' dataframe with one day per row,
                 and the following column hierarchy: (columns_to_sum, group_column)
                 Each day, we have the running sum of each columns_to_sum, for each group of contracts.
                 The first row is the earliest contract start date and the last row is the latest
                 contract start or contract end date.
        """

        cls.check_contracts_dates(contracts, date_start_col, date_end_exclusive_col)

        for c in columns_to_sum:
            if c not in contracts.columns:
                raise ValueError("missing column_to_sum: {}".format(c))
            if contracts[c].isnull().any():
                rows_with_nan_c = contracts[contracts[c].isnull()]
                raise ValueError("There are NaN values for column {} in these rows:\n{}".format(c, rows_with_nan_c))

        # keep only useful columns
        df = contracts[[date_start_col, date_end_exclusive_col] + columns_to_sum]
        # create start and end events for each contract, sorted chronologically
        events = cls.__contract_to_events(df, date_start_col, date_end_exclusive_col)

        # remove events after max_date if they are not wanted
        if max_date_exclusive is not None:
            events = events[events["event_date"] <= max_date_exclusive]

        # for each column to sum, replace the value by their "increment" (+X if contract starts; -X if contract ends)
        for c in columns_to_sum:
            events[c] = events.apply(lambda row: row[c] if row["event_type"] == "start" else -row[c], axis=1)

        # group events by day and sum the individual contract increments of columns_to_sum to have daily increments
        df_by_day = events.groupby(["event_date"]).sum()

        # add days that have no increment (with NA values), else the result can have gaps
        # new "NA" increments = no contract start or end event that day = increment is 0
        df_by_day = df_by_day.asfreq('D').fillna(0)

        # compute cumulative sums of daily increments to get daily totals
        portfolio = df_by_day.cumsum(axis=0)
        portfolio.index.name = "date"  # now values are not increments on an event_date but the total on this date

        return portfolio

    @staticmethod
    def get_portfolio_between_dates(portfolio, start_datetime, end_datetime_exclusive):
        """
        Adds or removes dates if needed.

        If additional dates are needed at the beginning, add these dates with 0s
        If additional dates needed at the end, copy the portfolio of the last date into the additional dates

        :param portfolio: the dataframe with the portfolio, must have a pandas.DatetimeIndex with frequency
        :param start_datetime:
        :param end_datetime_exclusive:
        :return:
        """

        df = portfolio.copy(deep=True)

        if not isinstance(df.index, pd.DatetimeIndex):
            raise TypeError("The index of daily_portfolio should be a pd.DatetimeIndex, but given {}"
                            .format(df.index.dtype))

        if df.index.freq is None:
            raise ValueError("portfolio.index needs to have a freq. "
                             "Maybe try to set one using df.index.inferred_freq")

        freq = df.index.freq

        # check that there is no missing value
        if not df.isnull().sum().sum() == 0:
            raise ValueError("daily_portfolio has NaN values.")

        if start_datetime is not None and df.index.min() > start_datetime:
            # add days with empty portfolio at the beginning
            df = df.append(pd.Series(name=start_datetime, dtype='object'))
            df.sort_index(inplace=True)  # put the new row first
            df = df.asfreq(freq).fillna(0)

        if end_datetime_exclusive is not None and df.index.max() < end_datetime_exclusive:
            # add days at the end, with the same portfolio as the last available day
            df = df.append(pd.Series(name=end_datetime_exclusive, dtype='object'))
            df.sort_index(inplace=True)  # make sure this new row is last
            df = df.asfreq(freq, method='ffill')

        # remove dates outside of desired range
        df = df[(df.index >= start_datetime) & (df.index < end_datetime_exclusive)]
        assert df.isnull().sum().sum() == 0  # check that there is no missing value

        return df

    @classmethod
    def forecast_using_trend(cls, portfolio_df, start_forecast_date, nb_days=14, past_days=100):
        """
        Forecast using exponential smoothing (Holt method) for the next nb_days
        The output has the same frequency as input portfolio_df.

        :param portfolio_df:
        :param start_forecast_date: when we stop the portfolio data and start forecasting
        :param nb_days: number of days after 'end_date' to forecast
        :param past_days: max number of days to use in the past used to make the forecast
                          (it is better to use only recent data)
        :return: pd.DataFrame (the forecast data)
        """

        try:
            from statsmodels.tsa.api import Holt
        except ImportError:
            raise ImportError("statsmodels is required is you want to use this function. "
                              "Try: pip install statsmodels>=0.12.0")

        if nb_days < 1:
            raise ValueError("nb_days should be at least 1, given {}".format(nb_days))

        if not isinstance(start_forecast_date, pd.Timestamp):
            raise ValueError("Expected a pandas.Timestamp for start_forecast_date, but got {} of type {}. "
                             "You can use pandas.to_datetime(...) before passing this argument."
                             .format(start_forecast_date, type(start_forecast_date)))

        if start_forecast_date > portfolio_df.index.max() + relativedelta(days=1):
            raise ValueError("Start forecast date ({}) more than 1 day after the latest portfolio information ({}). "
                             "Portfolio information given is too old."
                             .format(start_forecast_date, portfolio_df.index.max()))

        if not isinstance(portfolio_df.index, pd.DatetimeIndex):
            raise ValueError("portfolio_df should have a pandas.DatetimeIndex, but given {}"
                             .format(portfolio_df.index.dtype))

        if portfolio_df.index.freq is None:
            raise ValueError("Input portfolio_df must have a frequency. "
                             "Maybe try to set it using pandas.index.inferred_freq")

        # only keep portfolio data before the start_forecast_date
        freq = portfolio_df.index.freq
        tzinfo = portfolio_df.index.tzinfo  # can be None

        pf = portfolio_df
        pf = pf[pf.index <= start_forecast_date]

        end_forecast_date = TimezoneUtils.add_interval_to_day_dt(
            day_dt=start_forecast_date,
            interval=relativedelta(days=nb_days)
        )
        date_past_days_ago = TimezoneUtils.add_interval_to_day_dt(
            day_dt=start_forecast_date,
            interval=relativedelta(days=-past_days)
        )

        # only keep recent data to determine the trends
        pf = pf[pf.index >= date_past_days_ago]

        future_index = pd.date_range(
            start_forecast_date,
            end_forecast_date,
            freq=freq,
            name=pf.index.name,
            closed='left'
        )
        if tzinfo is not None:
            future_index = future_index.tz_convert(pf.index.tzinfo)

        # holt needs a basic integer index
        pf = pf.reset_index(drop=True)
        # forecast each column (all columns are measures)
        result = pf.apply(lambda x: Holt(x, initialization_method="estimated").fit().forecast(len(future_index)))
        result = result.round(1)
        result.index = future_index

        return result




import pandas as pd
import datetime
import pytz


class TimeSeries:

    @classmethod
    def align_timezone(cls, time_series: pd.Series, tzinfo: [str, pytz.timezone]):
        """
        Sometimes a time_series is of pandas type "object" just because the time-zone information
        is not well read initially.

        Example :
        time_series = a time_series with some times at timezone +01:00 (French winter time)
                      and others at timezone +02:00 (French summer time)
                      So its pandas dtype is "object"
        tz = pytz.timezone("Europe/Paris")
        We want to make sure the timezone information is tzinfo for each row and also for the series.

        :param time_series: a series with tz-aware date-times
        :param tzinfo: a str or a datetime.tzinfo
        :return: a DatetimeIndex of dtype: datetime[ns, tzinfo]
        """

        if not isinstance(time_series, pd.Series) and not isinstance(time_series, pd.DatetimeIndex):
            raise TypeError("parameter 'series' should be a pandas.Series")

        if isinstance(tzinfo, str):
            tzinfo = pytz.timezone(tzinfo)

        if not isinstance(tzinfo, datetime.tzinfo):
            raise TypeError("parameter 'tzinfo' should be of type str or datetime.tzinfo")

        result = time_series.map(lambda x: x.astimezone(tzinfo))
        # now all values in 'result' have the same type: pandas.Timestamp with the same tzinfo
        result = pd.DatetimeIndex(result)
        return result

    @classmethod
    def find_missing_and_extra_periods(
            cls,
            dti,
            expected_freq=None,
            expected_start_datetime=None,
            expected_end_datetime=None):
        """
        Check for missing and extra data points
        :param dti: a series of type DatetimeIndex, in ascending order and without duplicates or NaNs
        :param expected_freq: pandas formatted frequency. If None is given, will infer the frequency, taking the
            most common gap between 2 consecutive points.
        :param expected_start_datetime: a pandas.Datetime, if None is given, will take dti[0]
        :param expected_end_datetime: a pandas.Datetime, if None is given, will take dti[-1]
        :return: Missing and extra data points collapsed in periods
        """

        if not isinstance(dti, pd.DatetimeIndex):
            raise TypeError("parameter 'dti' should be a pandas.DatetimeIndex")

        if dti.isna().sum() != 0:
            raise ValueError("given dti has NaN values.")

        if dti.duplicated().sum() != 0:
            raise ValueError("given dti has duplicates.")

        if not dti.equals(dti.sort_values(ascending=True)):
            raise ValueError("given dti is not in ascending order")

        if len(dti) == 0:
            raise ValueError("given dti is empty")

        if expected_start_datetime is not None:
            if not isinstance(expected_start_datetime, pd.Timestamp):
                raise TypeError("expected_start_datetime must be a pandas.Datetime")
            start = expected_start_datetime
        else:
            start = dti[0]

        if expected_end_datetime is not None:
            if not isinstance(expected_end_datetime, pd.Timestamp):
                raise TypeError("expected_end_datetime must be a pandas.Datetime")
            end = expected_end_datetime
        else:
            end = dti[-1]

        if expected_freq is not None:
            if not isinstance(expected_freq, str) and not isinstance(expected_freq, pd.Timedelta):
                raise TypeError("expected_freq should be str or pd.Timedelta")
            freq = expected_freq

        else:
            # Infer frequency

            # compute the gap distribution
            gap_dist = (pd.Series(dti[1:]) - pd.Series(dti[:-1])).value_counts()
            # gap_dist =
            #    01:00:00    1181
            #    02:00:00     499
            #    03:00:00     180
            #    ....

            # take the most common gap and use it as expected_freq
            freq = gap_dist.index[0]  # this is a pd.Timedelta object
            assert isinstance(freq, pd.Timedelta)

        expected_index = pd.date_range(start, end, freq=freq)
        missing_points = expected_index.difference(dti)
        # group missing points together as "missing periods"
        missing_periods = cls.collapse_dt_series_into_periods(missing_points, freq=freq)
        extra_points = dti.difference(expected_index)

        return freq, missing_periods, extra_points

    @classmethod
    def collapse_dt_series_into_periods(
            cls,
            dti: pd.DatetimeIndex,
            freq: [str, pd.Timedelta]
    ):
        """
        This function does not work if freq < 1s
        :param freq:
        :param dti: DatetimeIndex, must be sorted
        :return:
        """

        assert isinstance(dti, pd.DatetimeIndex)
        assert isinstance(freq, str) or isinstance(freq, pd.Timedelta)
        if pd.to_timedelta(freq).total_seconds() < 1.0:
            raise ValueError("freq must be more than 1 second, but given {}".format(freq))

        if dti.shape[0] == 0:
            return []

        current_period_start = dti[0]
        periods_list = []
        for i in range(1, dti.shape[0]):
            if dti[i] <= dti[i-1]:
                raise ValueError("dti must be sorted and without duplicates!")

            if (dti[i] - dti[i-1]).total_seconds() % pd.to_timedelta(freq).total_seconds() != 0:
                raise ValueError("Timedelta between {} and {} is not a multiple of freq ({})."
                                 .format(dti[i-1], dti[i], freq))

            if pd.to_timedelta(freq) != dti[i] - dti[i-1]:  # End the current period and start a new one
                periods_list.append((current_period_start, dti[i-1]))
                current_period_start = dti[i]
        periods_list.append((current_period_start, dti[-1]))  # Don't forget last period
        return periods_list

    @staticmethod
    def interpolate_daily_to_sub_daily_data(
            df: pd.DataFrame,
            freq: [str, pd.Timedelta],
            tz: [str, datetime.tzinfo],
            index_name: str = 'time',
            method: str = 'ffill'):
        """
        Interpolate daily data in a dataframe (with a DatetimeIndex) to sub-daily data using a given method.
        :param df: pd.DataFrame
        :param freq: a frequency < 'D' (e.g. 'H', '30min', '15min', etc)
        :param tz: the time zone (None not accepted because important)
        :param index_name: name to give to the new index. Usually going from 'date' to 'time'.
        :param method: how are data interpolated between two consecutive dates (e.g. 'ffill', 'linear', etc)
        :return: pd.DataFrame
        """

        assert type(df.index) == pd.DatetimeIndex
        assert pd.to_timedelta(freq).total_seconds() <= 24*60*60

        if df.index.tzinfo is None:
            df.index = pd.to_datetime(df.index).tz_localize(tz)
        else:
            df.index = pd.to_datetime(df.index).tz_convert(tz)

        new_end_date = df.index.max() + datetime.timedelta(days=1)
        extra_row = df.iloc[[-1]].reindex([new_end_date])

        result = df.append(extra_row, ignore_index=False)
        result = result.resample(freq).interpolate(method=method)
        result = result.drop(new_end_date)
        result.index.name = index_name

        return result



# Let users know if they're missing any of our hard dependencies
hard_dependencies = ("pandas",)
# note that we also need python-dateutil and pytz but pandas already depends on them, so importing pandas is enough
missing_dependencies = []

for dependency in hard_dependencies:
    try:
        __import__(dependency)
    except ImportError as e:
        missing_dependencies.append(dependency)

if len(missing_dependencies) > 0:
    raise ImportError(
        "Unable to import required dependencies:\n" + "\n".join(missing_dependencies)
    )
del hard_dependencies, dependency, missing_dependencies

# import some ubclasses here so users can use for instance :
# 'enda.Contracts' without knowing the internal structure.
# Do not import classes that need a specific packages like "H2OModel".
from enda.contracts import (Contracts)
from enda.feature_engineering.datetime_features import (DatetimeFeature)
from enda.timeseries import (TimeSeries)
from enda.backtesting import (BackTesting)



import pandas as pd
import datetime
from enda.feature_engineering.datetime_features import DatetimeFeature

try:
    import unidecode
    from jours_feries_france import JoursFeries
    from vacances_scolaires_france import SchoolHolidayDates
except ImportError:
    raise ImportError("unidecode, jours_feries_france, vacances_scolaires_france are required if you want to use "
                      "enda's FrenchHolidays and FrenchCalendar classes. "
                      "Try: pip install jours-feries-france vacances-scolaires-france Unidecode")


class FrenchHolidays:

    @staticmethod
    def get_public_holidays(year_list=None, orientation='rows'):

        if year_list is None:
            year_list = range(2000, 2051)

        result = pd.DataFrame()
        for year in year_list:
            try:
                res = JoursFeries.for_year(year)
                df_res = pd.DataFrame.from_dict(res, orient='index')
                df_res.index = df_res.index.map(lambda x: unidecode.unidecode(x))
                result = result.append(df_res.T, ignore_index=True)
            except Exception as e:
                raise ValueError("Missing french public holidays : {}".format(e))

        if orientation is not 'columns':
            result = result.stack().reset_index(level=0, drop=True)
            result.index.name = 'nom_jour_ferie'
            result = result.to_frame('date')
            result = result.reset_index(drop=False)
            result = result[['date', 'nom_jour_ferie']]

        return result

    @staticmethod
    def get_school_holidays(year_list=None):

        if year_list is None:
            current_year = datetime.datetime.utcnow().year
            year_list = range(2000, current_year + 1)

        d = SchoolHolidayDates()
        result = pd.DataFrame()
        for year in year_list:
            try:
                res = d.holidays_for_year(year)
                df_res = pd.DataFrame.from_dict(res, orient='index')
                df_res['nom_vacances'] = df_res['nom_vacances'].map(lambda x: unidecode.unidecode(x))
                df_res = df_res.reset_index(drop=True)
                result = result.append(df_res, ignore_index=True)
            except Exception as e:
                raise ValueError("Missing french school holidays : {}".format(e))

        return result


class Calendar:

    def __init__(self, country='FR'):
        self.country = country

    def get_french_lockdown(self):
        """
        Return a dataframe from 2000-01-01 to 2050-12-25 indicating for each day if national lockdown was ongoing.
        So far, the main lockdown period goes from 2020-03-17 to 2020-05-11.
        """

        if self.country is not 'FR':
            raise NotImplementedError("Public holidays in {} unknown".format(self.country))

        start_lockdown_date = pd.to_datetime('2020-03-17')
        end_lockdown_date = pd.to_datetime('2020-05-11')
        lockdown_period = pd.date_range(start_lockdown_date, end_lockdown_date)

        df_lockdown = pd.DataFrame(index=lockdown_period, columns=['lockdown'], data=1)
        df_lockdown.index.name = 'date'

        result = df_lockdown.reindex(pd.date_range('2000-01-01', '2050-12-25'))
        result = result.fillna(0)

        return result

    def get_public_holidays(self):
        """
        Return a dataframe from 2000-01-01 to 2050-12-25 indicating for each day
        whether it is a public holiday (denoted by a 1) or not (denoted by a 0)
        """

        if self.country is 'FR':
            public_holidays = FrenchHolidays.get_public_holidays()
        else:
            raise NotImplementedError("Public holidays in {} unknown".format(self.country))

        public_holidays = public_holidays.set_index('date')
        public_holidays.index = pd.to_datetime(public_holidays.index)
        public_holidays = public_holidays[~public_holidays.index.duplicated(keep='first')]  # 2008-05-01

        public_holidays['public_holiday'] = 1
        public_holidays = public_holidays.asfreq('D')
        public_holidays = public_holidays.fillna(0)

        return public_holidays[['public_holiday']]

    def get_school_holidays(self):
        """
        Return a dataframe from 2000-01-01 to as far as possible (2021-08-29) indicating for each day
        the number of school areas (zone A, B et C) in vacation (either 0, 1, 2 or 3)
        """

        if self.country is 'FR':
            school_holidays = FrenchHolidays.get_school_holidays()
        else:
            raise NotImplementedError("School holidays in {} unknown".format(self.country))

        school_holidays = school_holidays.set_index('date')
        school_holidays.index = pd.to_datetime(school_holidays.index)
        school_holidays = school_holidays.drop('nom_vacances', 1)

        school_holidays['nb_school_areas_off'] = school_holidays.sum(axis=1)
        school_holidays = school_holidays.asfreq('D')
        school_holidays = school_holidays.fillna(0)

        return school_holidays[['nb_school_areas_off']]

    def get_extra_long_weekend(self):
        """
        Return a dataframe from 2000-01-01 to 2050-12-25 indicating for each day
        - if the previous (resp. the next day) is a public holiday
        AND
        - if the current day is a tuesday (resp. a thursday)
        If both conditions are fulfilled then the day is denoted by a 1 (0 otherwise)
        """

        public_holidays = self.get_public_holidays()
        public_holidays = DatetimeFeature.split_datetime(public_holidays, split_list=['dayofweek'], index=True)

        public_holidays['is_yesterday_day_off'] = public_holidays['public_holiday'].shift()
        public_holidays['is_tomorrow_day_off'] = public_holidays['public_holiday'].shift(-1)
        public_holidays['extra_long_weekend'] = 0

        mondays = public_holidays[public_holidays['dayofweek'] == 0]
        mondays_off = mondays[mondays['is_tomorrow_day_off'] == 1].index

        fridays = public_holidays[public_holidays['dayofweek'] == 4]
        fridays_off = fridays[fridays['is_yesterday_day_off'] == 1].index

        extra_long_weekend_index = mondays_off.append(fridays_off)
        extra_long_weekend_index = sorted(extra_long_weekend_index)

        public_holidays.loc[extra_long_weekend_index, 'extra_long_weekend'] = 1

        return public_holidays[['extra_long_weekend']]

    @staticmethod
    def interpolate_daily_to_subdaily_data(df, freq, method='ffill', tz='Europe/Paris'):
        """
        Interpolate daily data in a dataframe (with a DatetimeIndex) to subdaily data using a given method.
        :param df: pd.DataFrame
        :param freq: a frequency < 'D' (e.g. 'H', '30min', '15min', etc)
        :param method: how are data interpolated between two consecutive dates (e.g. 'ffill', 'linear', etc)
        :param tz: timezone ('Europe/Paris')
        :return: pd.DataFrame
        """
        if df.index.tzinfo is None:
            df.index = pd.to_datetime(df.index).tz_localize(tz)

        new_end_date = df.index.max() + datetime.timedelta(days=1)
        extra_row = df.iloc[[-1]].reindex([new_end_date])

        result = df.append(extra_row, ignore_index=False)
        result = result.resample(freq).interpolate(method=method)
        result = result.drop(new_end_date)

        result.index.name = 'time'

        return result

    @staticmethod
    def get_french_special_days(freq='30min'):

        lockdown = Calendar().get_french_lockdown()
        public_holidays = Calendar().get_public_holidays()
        school_holidays = Calendar().get_school_holidays()
        extra_long_weekend = Calendar().get_extra_long_weekend()

        result = pd.concat([lockdown, public_holidays, school_holidays, extra_long_weekend], 1, 'inner')
        result = Calendar().interpolate_daily_to_subdaily_data(result, freq=freq)

        return result



import abc
import pandas
from collections import OrderedDict
import typing


class ModelInterface(metaclass=abc.ABCMeta):
    """
    This interface represents a simple machine learning model with some universal functions.
    We require these functions :
        train : train the model
        predict : predict using the model

    To save and load instances of a class, use tools like pickle or joblib
    (see information for instance here: https://scikit-learn.org/stable/modules/model_persistence.html).

    This interface is useful to create more advanced models based on these building blocks :
        NormalizedModel : uses one of the inputs as a 'normalization variable' instead of a 'training feature'.
        StackingModel : combines several models to create a more robust model (cross-algorithm 'ensemble' method).
        ModelWithFallback : a model that can cope with missing input in a robust way.

    See tutorials about Python interfaces for instance here https://realpython.com/python-interface/ .
    """

    @classmethod
    def __subclasshook__(cls, subclass):
        return (
            hasattr(subclass, 'train') and callable(subclass.train) and
            hasattr(subclass, 'predict') and callable(subclass.predict))

    @abc.abstractmethod
    def train(self, df: pandas.DataFrame, target_col: str):
        """Trains the model using the given data."""
        raise NotImplementedError

    @abc.abstractmethod
    def predict(self, df: pandas.DataFrame, target_col: str) -> pandas.DataFrame:
        """ Predicts and returns a dataframe with just 1 column: target_col_name """
        raise NotImplementedError


class NormalizedModel(ModelInterface):

    def __init__(self,
                 normalized_model: ModelInterface,
                 target_col: str,
                 normalization_col: str,
                 columns_to_normalize: [typing.Iterable[str], None] = None):
        """
        A model that uses one of the inputs as a linear 'normalization variable' instead of a 'training feature'.
        The prediction is:
        predict(X) = X[normalization_col] * predict(X without normalization_col)

        The predict(X without normalization_col) is made by another underlying model.

        :param normalized_model: the model that will perform predict(X without normalization_col)
        :param normalization_col: name of the column used as a multiplier and not as a feature.
        :param columns_to_normalize: (optional) columns in 'X without normalization_col' that must be divided
        by normalization_col for the the underlying model to train/predict correctly.
        """

        if not issubclass(type(normalized_model), ModelInterface):
            raise TypeError("model's type '{}' should implement enda.ModelInterface".format(type(normalized_model)))

        if columns_to_normalize and normalization_col in columns_to_normalize:
            raise ValueError("normalisation_col '{}'should not be in columns_to_normalize {}"
                             .format(normalization_col, columns_to_normalize))

        self.normalized_model = normalized_model
        self.target_col = target_col
        self.normalisation_col = normalization_col
        self.columns_to_normalize = columns_to_normalize

    def check_normalization_col(self, df: pandas.DataFrame):
        zeros_df = df[df[self.normalisation_col] <= 0]
        if not zeros_df.empty:
            raise ValueError("Normalisation_col '{}' : zeros found\n{}".format(self.normalisation_col, zeros_df))

    def normalize(self, df: pandas.DataFrame):
        self.check_normalization_col(df)
        df_norm = df.copy(deep=True)

        if self.columns_to_normalize:
            for c in df.columns:
                if c in self.columns_to_normalize:
                    df_norm[c] = df_norm[c]/df[self.normalisation_col]

        # always normalize the target if it is in the df (present in train mode, not in predict mode)
        if self.target_col in df.columns:
            df_norm[self.target_col] = df_norm[self.target_col]/df[self.normalisation_col]

        df_norm.drop(columns=self.normalisation_col, inplace=True)
        return df_norm

    def train(self, df: pandas.DataFrame, target_col: str = None, drop_where_normalization_under_zero: bool = False):
        if target_col and self.target_col != target_col:
            raise ValueError("target should be None or {}, but given: {}".format(self.target_col, target_col))

        if drop_where_normalization_under_zero:
            df = df.loc[df[self.normalisation_col] > 0, :]
        df_norm = self.normalize(df)
        self.normalized_model.train(df_norm, self.target_col)

    def predict(self, df: pandas.DataFrame, target_col: str = None):
        if target_col and self.target_col != target_col:
            raise ValueError("target should be None or '{}', but given: '{}'".format(self.target_col, target_col))

        df_norm = self.normalize(df)  # error out if any value of normalization_col is <= 0
        predict_norm = self.normalized_model.predict(df_norm, self.target_col)

        if (predict_norm.index != df.index).any():
            raise ValueError("prediction must have the same index as given df. "
                             "Check that for self.normalized_model, the method 'predict' conserves index.")

        predict = predict_norm.multiply(df[self.normalisation_col], axis='index')
        return predict


class StackingModel(ModelInterface):

    def __init__(self,
                 base_models: typing.Mapping[str, ModelInterface],
                 final_model: ModelInterface):
        """
        This class serves the same purpose as the Scikit-Learn "Stacking Regressor". However since we work on
        time-series, we need fine control on which data is passed to train the base_models before training
        the final_model.


        :param base_models: a dict of {model_id -> model}, each model must implement enda.ModelInterface.
        :param final_model: the model used for stacking, must also implement enda.ModelInterface
        """

        if len(base_models) <= 1:
            raise ValueError("At least 2 base_models are required, but given: {}".format(len(base_models)))

        # We store models in an ordered dict to make sure we always iterate over them in the same order
        self.base_models = OrderedDict()
        for model_id in sorted(base_models.keys()):
            self.base_models[model_id] = base_models[model_id]

        self.final_model = final_model

    def train(self, df: pandas.DataFrame, target_col: str, base_stack_split_pct: float = 0.05):

        # training stacking will temporarily train single models on part of the data,
        # so it must be done before training the actual single models
        self._train_final_model(df, target_col, base_stack_split_pct)

        # re-train base models with the full dataset
        self._train_base_models(df, target_col)

    def _train_final_model(self, df, target_col, base_stack_split_pct):
        """
        Trains the final model used for stacking.

        (Temporarily) train the single models with a subset of the data,
        then apply them on the rest of the data. Use this to train the stacking model.
        """

        # split the training frame : ,
        split_int = int(df.shape[0] * (1-base_stack_split_pct))
        split_idx = df.index[split_int]

        df_base_models = df[df.index < split_idx]  # one part to train the base models
        df_stacking = df[df.index >= split_idx]  # the other to train the stacking model

        if df_base_models.shape[0] == 0 or df_stacking.shape[0] == 0:
            raise ValueError("The split gave an empty train set for the base models or the stacking model. Change"
                             "parameter 'base_stack_split_pct' (given {}) or provide a larger training set."
                             .format(base_stack_split_pct))

        self._train_base_models(df_base_models, target_col)

        # make predictions with these temporary base_models, without the target column
        base_model_predictions = self._predict_base_models(df_stacking.drop(columns=[target_col]), target_col)

        # add the target back, to train the stacking model
        base_model_predictions[target_col] = df_stacking[target_col]

        self.final_model.train(base_model_predictions, target_col)

    def _predict_base_models(self, df, target_col):
        """
        :return: a Dataframe with the prediction of each base model in each column.
        """

        model_dfs = []
        for model_id, model in self.base_models.items():
            model_predict = model.predict(df, target_col)

            if (model_predict.index != df.index).any():
                raise ValueError("prediction must have the same index as given df. "
                                 "Check that for model with id '{}', the method 'predict' conserves index."
                                 .format(model_id))

            model_predict.rename(columns={target_col: model_id}, inplace=True)
            model_dfs.append(model_predict)

        predict_df = pandas.concat(model_dfs, axis=1, join='outer')

        if predict_df.shape[0] != df.shape[0]:
            raise ValueError("Given {} values to predict, but predicted {}"
                             .format(df.shape[0], predict_df.shape[0]))

        return predict_df

    def _train_base_models(self, df, target_col):
        # train base models
        for model_id, model in self.base_models.items():
            model.train(df, target_col)

    def predict(self, df: pandas.DataFrame, target_col: str):
        base_model_predictions = self._predict_base_models(df, target_col)
        prediction = self.final_model.predict(base_model_predictions, target_col)

        if (prediction.index != df.index).any():
            raise ValueError("prediction must have the same index as given df. "
                             "Check that self.final_model.predict conserves index.")

        return prediction


class ModelWithFallback(ModelInterface):
    """
    This models allows to make a prediction even when some important input is missing.

    In order to deal with missing values, it is common practice to replace None/NA with some meaningful value like
    the mean or the median of the values found in the train set for this feature.
    However this is problematic when the missing variable has a significant impact on the prediction.

    Instead, this model trains 2 underlying models: a 'model_with' and a 'model_without' the column that can be missing.
    When predicting, it will use the 'model_with' for inputs with the column present and the 'model_without' for the
    others.
    """

    def __init__(self,
                 resilient_column: str,
                 model_with: ModelInterface,
                 model_without: ModelInterface):
        """ Provide 2 different raw models ready to be trained. """

        if model_with is model_without:  # check identity
            raise AttributeError("model_with and model_without must be different objects. If you want the same base "
                                 "model, you can use copying tools (like copy.deepcopy()) to duplicate the raw model.")

        self.resilient_column = resilient_column
        self.model_with = model_with
        self.model_without = model_without

    def train(self, df: pandas.DataFrame, target_col: str):
        """
        Trains the two models : model_with and model_without the 'column_name'
        """

        # only train the "model_with" where column_name is present (not NA)
        self.model_with.train(df.dropna(subset=[self.resilient_column]), target_col)
        # train the "model_without" column_name
        self.model_without.train(df.drop(columns=[self.resilient_column]), target_col)

    def predict_both(self, df: pandas.DataFrame, target_col: str):
        df_with = df[df[self.resilient_column].notna()]
        predict_with = self.model_with.predict(df_with, target_col)
        if predict_with.shape[0] != df.shape[0]:
            # Add missing rows with NaN prediction
            predict_with = predict_with.reindex(df.index)

        df_without = df.drop(columns=[self.resilient_column])
        predict_without = self.model_without.predict(df_without, target_col)

        if (predict_with.index != df.index).any():
            raise ValueError("predict_with must have the same index as given df. "
                             "Check that self.model_with.predict conserves index.")

        if (predict_without.index != df.index).any():
            raise ValueError("predict_without must have the same index as given df. "
                             "Check that self.predict_without.predict conserves index.")

        return predict_with, predict_without

    def predict(self, df: pandas.DataFrame, target_col: str) -> pandas.DataFrame:
        predict_with, predict_without = self.predict_both(df, target_col)

        # keep prediction with column_name when available, else take the prediction of the model without it
        result = predict_with[target_col].fillna(predict_without[target_col])  # pandas series
        result = result.to_frame(target_col)
        return result



import pandas as pd
from enda.timezone_utils import TimezoneUtils
from dateutil.relativedelta import relativedelta


class BackTesting:
    """
    A class to help with back-testing on algorithms.
    """

    @staticmethod
    def yield_train_test(
            df,
            start_eval_datetime,
            days_between_trains,
            gap_days_between_train_and_eval=0
    ):
        """
        Returns pairs of (train set, test set) to perform back-testing on the data

        :param df: the dataset, a pandas.DataFrame with a pandas.DatetimeIndex
        :param start_eval_datetime: the beginning of the first eval, with same timezone as df.index
        :param days_between_trains: number of days between two train sets, it is also the duration of each test set
        :param gap_days_between_train_and_eval: optional, represents the time gap between the moment when
                                                things occurred and when we have the data available for training.
                                                (typically a few days or weeks)
        :return: a generator of (train set, test set) pairs
        """

        if not isinstance(df.index, pd.DatetimeIndex):
            raise ValueError("df must have a pandas.DatetimeIndex index")

        if str(df.index.tz) != str(start_eval_datetime.tzinfo):
            raise ValueError("df.index (tzinfo={}) and start_eval_datetime (tzinfo={}) must have the same "
                             "tzinfo.".format(df.index.tz, start_eval_datetime.tzinfo))

        if start_eval_datetime <= df.index.min():
            raise ValueError("start_eval_datetime ({}) must be after the beginning of df ({})"
                             .format(start_eval_datetime, df.index.min()))

        if not (start_eval_datetime.hour == start_eval_datetime.minute
                == start_eval_datetime.second == start_eval_datetime.microsecond == 0):
            raise ValueError("start_eval_datetime must be datetime with only years, months or days (not more precise),"
                             " but given: {}, {}".format(type(start_eval_datetime), start_eval_datetime))

        # initialize date-times: end_train, start_test and end_test
        if gap_days_between_train_and_eval > 0:
            end_train = TimezoneUtils.add_interval_to_day_dt(
                start_eval_datetime,
                relativedelta(days=-gap_days_between_train_and_eval)
            )
        else:
            end_train = start_eval_datetime
        train_interval = relativedelta(days=days_between_trains)
        start_test = start_eval_datetime
        end_test = TimezoneUtils.add_interval_to_day_dt(start_test, train_interval)

        # go through the dataset and yield pairs of (train set, test set)
        while start_test < df.index.max():
            yield df[df.index < end_train], df[(df.index >= start_test) & (df.index < end_test)]

            end_train = TimezoneUtils.add_interval_to_day_dt(end_train, train_interval)
            start_test = TimezoneUtils.add_interval_to_day_dt(start_test, train_interval)
            end_test = TimezoneUtils.add_interval_to_day_dt(end_test, train_interval)



import datetime
import pandas as pd
import numpy as np

# TODO : unittest


class DatetimeFeature:

    @staticmethod
    def split_datetime(df, split_list=None, index=True, colname=None):
        """
        Split a specific datetime column or datetime index into different date and time attributes (given by split list)
        Return the dataframe df with the new columns.
        :param df: pd.DataFrame
        :param split_list: attributes in ['minute', 'minuteofday', 'hour', 'day', 'month', 'year', 'dayofweek',
         'weekofyear', 'dayofyear']
        :param index: Bool (True if working on a DatetimeIndex, False if working on a column)
        :param colname: str (only if index=False)
        :return: pd.DataFrame with new columns (split_list)
        """

        if index is False and colname is None:
            raise ValueError("To split a datetime column, please specify the name of the column (argument colname)\n"
                             "To split a datetime index, please specify the argument index=True")

        if index is True and colname is not None:
            raise ValueError("You can split either your index or a given column but not both at the same time")

        split_list_implemented = ['minute', 'minuteofday', 'hour', 'day', 'month', 'year', 'dayofweek', 'weekofyear',
                                  'dayofyear']
        if split_list is None:
            split_list = split_list_implemented
        else:
            if not all(split in split_list_implemented for split in split_list):
                extra_split = list(set(split_list) - set(split_list_implemented))
                raise NotImplementedError('Split {} not in {}'.format(extra_split, split_list_implemented))

        df_split = pd.DataFrame(index=df.index)
        if index:
            if not isinstance(df.index, pd.DatetimeIndex):
                raise TypeError("To split a datetime index, please provide a DatetimeIndex in your dataframe")

            for split in split_list:
                if split is 'weekofyear':  # To avoid a FutureWarning
                    df_split[split] = df.index.isocalendar().week
                elif split is 'minuteofday':
                    df_split[split] = df.index.hour * 60 + df.index.minute
                else:
                    df_split[split] = df.index.__getattribute__(split)

        else:
            if colname not in list(df.select_dtypes(include=[np.datetime64])):
                raise TypeError("{} is not a datetime column : {}".format(colname, df[[colname]].dtypes))

            for split in split_list:
                if split is 'weekofyear':  # To avoid a FutureWarning
                    df_split[split] = df[colname].dt.isocalendar().week
                else:
                    df_split[split] = df[colname].dt.__getattribute__(split)

        result = pd.concat([df, df_split], 1, 'inner')

        return result

    @staticmethod
    def get_nb_hours_in_day(d):
        if d.tzinfo is None:
            raise AttributeError("Please provide a timezone information to your datetime")

        return len(pd.date_range(d.date(), d.date() + datetime.timedelta(days=1), tz=d.tzinfo, freq='H', closed='left'))

    @staticmethod
    def daylight_saving_time_dates():
        """
        Return a pd.Dataframe with
        - as index : the dates when daylight saving time starts or ends
        - as column : the number of hour in this particular day (23 or 25)
        Example :
                                   nb_hour
        1995-03-26 00:00:00+01:00       23
        1995-09-24 00:00:00+02:00       25
        1996-03-31 00:00:00+01:00       23
                                       ...
        2027-10-31 00:00:00+02:00       25
        2028-03-26 00:00:00+01:00       23
        """

        df = pd.DataFrame(index=pd.date_range('1995-01-01', '2030-01-01', freq='H', tz='Europe/Paris', closed='left'))
        df['nb_hour'] = df.index.hour
        df_by_day = df.resample('D').count()
        df_daylight_saving_time_dates = df_by_day[df_by_day['nb_hour'] != 24]
        return df_daylight_saving_time_dates

    @staticmethod
    def encode_cyclic_datetime(d):
        """
        Get the cyclic properties of a datetime, represented as points on the unit circle.
        :param d: pandas datetime object
        :return: pd.DataFrame of sine and cosine
        """

        days_in_year = 365 if not d.is_leap_year else 366
        days_in_month = d.days_in_month

        month = d.month - 1
        dayofmonth = d.day - 1
        dayofyear = d.dayofyear - 1

        result = {
            'minute_cos': [np.cos(2 * np.pi * d.minute / 60)],
            'minute_sin': [np.sin(2 * np.pi * d.minute / 60)],

            'minuteofday_cos': [np.cos(2 * np.pi * (d.hour * 60 + d.minute) / (24 * 60))],
            'minuteofday_sin': [np.sin(2 * np.pi * (d.hour * 60 + d.minute) / (24 * 60))],

            'hour_cos': [np.cos(2 * np.pi * d.hour / 24)],
            'hour_sin': [np.sin(2 * np.pi * d.hour / 24)],

            'day_cos': [np.cos(2 * np.pi * dayofmonth / days_in_month)],
            'day_sin': [np.sin(2 * np.pi * dayofmonth / days_in_month)],

            'month_cos': [np.cos(2 * np.pi * month / 12)],
            'month_sin': [np.sin(2 * np.pi * month / 12)],

            'dayofweek_cos': [np.cos(2 * np.pi * d.dayofweek / 7)],
            'dayofweek_sin': [np.sin(2 * np.pi * d.dayofweek / 7)],

            'dayofyear_cos': [np.cos(2 * np.pi * dayofyear / days_in_year)],
            'dayofyear_sin': [np.sin(2 * np.pi * dayofyear / days_in_year)],

        }

        result = pd.DataFrame.from_dict(data=result)
        result.index = [d]

        return result

    @staticmethod
    def encode_cyclic_datetime_index(df, split_list=None):
        """
        Split and encode a datetime index into different date and time attributes (given by split list).
        Encoding method : for each attribute, cosinus and sinus are provided.
        Return the dataframe df with the new columns.
        :param df: pd.DataFrame
        :param split_list: attributes in ['hour', 'day', 'month', 'dayofweek', 'dayofyear']
        :return: pd.DataFrame with new columns ['hour_cos', 'hour_sin', 'day_cos', ...]
        """

        max_dict = {'minute': 60, 'hour': 24, 'minuteofday': 1440, 'month': 12, 'dayofweek': 7}
        split_list_implemented = ['minute', 'minuteofday', 'hour', 'day', 'month', 'dayofweek', 'dayofyear']
        if split_list is None:
            split_list = split_list_implemented
        else:
            if not all(split in split_list_implemented for split in split_list):
                extra_split = list(set(split_list) - set(split_list_implemented))
                raise NotImplementedError('Split {} not in {}'.format(extra_split, split_list_implemented))

        df_split = pd.DataFrame(index=df.index)
        if not isinstance(df.index, pd.DatetimeIndex):
            raise TypeError("To split a datetime index, please provide a DatetimeIndex in your dataframe")

        for split in split_list:
            if split is 'minuteofday':
                df_split[split] = df.index.hour * 60 + df.index.minute
            else:
                df_split[split] = df.index.__getattribute__(split)

            if split in max_dict.keys():
                df_split['{}_max'.format(split)] = max_dict[split]

            if split is 'dayofyear':
                df_split['{}_max'.format(split)] = df_split.index.is_leap_year
                df_split['{}_max'.format(split)] = df_split['{}_max'.format(split)].replace({True: 366, False: 365})
                df_split['{}_max'.format(split)] = df_split['{}_max'.format(split)].astype(int)

            if split is 'day':
                df_split['{}_max'.format(split)] = df_split.index.days_in_month

            if split in ['month', 'day', 'dayofyear']:
                df_split[split] = df_split[split] - 1

            df_split['{}_{}'.format(split, 'cos')] = np.cos(
                (2 * np.pi * df_split[split]) / df_split['{}_max'.format(split)])
            df_split['{}_{}'.format(split, 'sin')] = np.sin(
                (2 * np.pi * df_split[split]) / df_split['{}_max'.format(split)])

            df_split = df_split.drop(split, 1)
            df_split = df_split.drop('{}_max'.format(split), 1)

        result = pd.concat([df, df_split], 1, 'inner')

        return result

    @staticmethod
    def get_datetime_features(df):
        result = DatetimeFeature.encode_cyclic_datetime_index(df)
        result = DatetimeFeature.split_datetime(result)
        return result



import os
import h2o
import time
import shutil
import pandas as pd
from enda.models import ModelInterface

try:
    from h2o.estimators import H2OXGBoostEstimator
    from h2o.estimators import H2OGradientBoostingEstimator
    from h2o.estimators import H2ORandomForestEstimator
    from h2o.estimators import H2ODeepLearningEstimator
    from h2o.estimators import H2OGeneralizedLinearEstimator
    from h2o.automl import H2OAutoML
    from h2o.grid.grid_search import H2OGridSearch
except ImportError:
    raise ImportError("h2o is required is you want to use this enda's H2OModel. Try: pip install h2o>=3.32.0.3")


class H2OModel(ModelInterface):

    def __init__(self, algo_name, model_id, algo_param_dict, target, model_path=None, verbose=False, logger=None,
                 seed=1234):

        self.algo_name = algo_name
        self.model_id = model_id
        self.algo_param_dict = algo_param_dict.copy()
        self.target = target
        self.model_path = model_path
        self.algo = None
        self.best_model = None
        self.verbose = verbose
        if self.verbose:
            if not logger:
                raise ValueError("When verbose=True, a logger must be provided")
        self.logger = logger

        algo_implemented_list = ['glm', 'gbm', 'xgboost', 'randomforest', 'deeplearning', 'automl']
        if self.algo_name not in algo_implemented_list:
            raise NotImplementedError("Algo {} not in {}".format(self.algo_name, algo_implemented_list))

        if 'glm' in self.algo_name.lower():
            if 'intercept' in self.algo_param_dict:
                intercept = self.algo_param_dict['intercept']
                del self.algo_param_dict['intercept']
                self.algo = H2OGeneralizedLinearEstimator(seed=seed, standardize=False, intercept=intercept)
            else:
                self.algo = H2OGeneralizedLinearEstimator(seed=seed, standardize=False)

        if 'gbm' in self.algo_name.lower():
            self.algo = H2OGradientBoostingEstimator(seed=seed)

        if 'xgboost' in self.algo_name.lower():
            self.algo = H2OXGBoostEstimator(seed=seed)

        if 'randomforest' in self.algo_name.lower():
            self.algo = H2ORandomForestEstimator(seed=seed)

        if 'deeplearning' in self.algo_name.lower():
            self.algo = H2ODeepLearningEstimator(seed=seed)

        if 'automl' in self.algo_name.lower():
            self.algo = H2OAutoML(seed=seed)

    def train(self, df: pd.DataFrame, target_col: str,
              validation_frame=None, search_criteria=None, nfolds=0, score_performance='rmse'):
        if target_col != self.target:
            raise ValueError

        self.grid_search(df, validation_frame, search_criteria, nfolds, score_performance)

    def grid_search(self, training_frame, validation_frame=None, search_criteria=None, nfolds=0,
                    score_performance='rmse'):
        """
        Allow grid search using self.algo_param_dict as hyperparameter space.
        By default the research is done by sampling uniformly 10 sets from the set of all possible hyperparameter value
        combinations.
        The best model is then the one showing the best 'score_performance' (RMSE by default) on the validation_frame.
        """

        h2o.no_progress()

        # clean h2o from any previous grid-search on this model,
        # else re-training using grid-search will error out
        h2o.remove(self.model_id)
        self.best_model = None

        if search_criteria is None:
            search_criteria = {'strategy': 'RandomDiscrete', 'max_models': 20, 'seed': 1234}

        train = h2o.H2OFrame(training_frame)
        valid = h2o.H2OFrame(validation_frame) if validation_frame is not None else None
        test = validation_frame.copy() if validation_frame is not None else training_frame.copy()

        perf_on_valid = True if validation_frame is not None else False

        x = list(training_frame.drop(self.target, 1))
        y = self.target

        grid = H2OGridSearch(model=self.algo, grid_id=self.model_id, hyper_params=self.algo_param_dict,
                             search_criteria=search_criteria)

        if self.verbose:
            self.logger.info("Start learning -- {} -- from {} to {}"
                             .format(self.model_id, training_frame.index.min(), training_frame.index.max()))

        start = time.time()
        # h2o.show_progress()

        grid.train(x=x, y=y, training_frame=train, validation_frame=valid, nfolds=nfolds)

        end = time.time()
        time_training = round((end - start) / 60.0, 2)

        grid_perf = grid.get_grid(sort_by=score_performance, decreasing=False)
        best_model = grid_perf.models[0]

        self.best_model = best_model

        if self.verbose:
            eval_df = self.predict(test)
            eval_df = pd.concat([eval_df, test[self.target]], 1, 'inner')
            # scoring = Scoring(eval_df, self.target, [self.model_id])
            # mape = scoring.mean_absolute_percentage_error().values[0]
            # bias = scoring.mean_error().values[0]
            print(grid_perf)
            print(best_model.varimp(use_pandas=True))
            print(best_model.model_performance(valid=perf_on_valid))
            # print("MAPE : {}".format(mape))
            # print("Bias : {}".format(bias))

        # LOGGER.info("End learning -- {} -- it took {} minutes".format(self.model_id, time_training))

        return best_model

    def predict(self, df: pd.DataFrame, target_col: str):

        h2o.no_progress()

        if self.best_model is None:
            raise ValueError("Prediction impossible, no model in memory, please use H2OModel.grid_search() "
                             "before trying to predict")

        testing_frame = df
        if self.target in df.columns:
            testing_frame = testing_frame.drop(self.target, 1)

        #LOGGER.info("Predicting -- {} -- from {} to {}"
        #    .format(self.model_id, testing_frame.index.min(), testing_frame.index.max()))

        test = h2o.H2OFrame(df)
        forecast = self.best_model.predict(test)
        forecast = forecast.as_data_frame()
        forecast = forecast.rename(columns={'predict': self.model_id})
        forecast.index = testing_frame.index

        return forecast

    @staticmethod
    def free_memory():
        h2o.remove_all()

    def save(self, tmp_dir=None):

        if self.best_model is None:
            raise ValueError("Saving impossible, no model in memory, please use H2OModel.grid_search() "
                             "before trying to save a model")

        model_path_tmp = os.path.join(tmp_dir, "tmp")
        if not os.path.exists(model_path_tmp):
            os.makedirs(model_path_tmp)

        model_path_from_h2o = h2o.save_model(model=self.best_model, path=model_path_tmp, force=True)
        # file_existed = os.path.exists(self.model_path)
        # message_log = 'overwritten' if file_existed else 'first time'
        shutil.move(model_path_from_h2o, self.model_path)
        # LOGGER.info("Model saved ({}) : {}".format(message_log, self.model_path))

    def load(self):
        # LOGGER.info("Loading model : {}".format(self.model_path))
        uploaded_model = h2o.upload_model(self.model_path)
        self.best_model = uploaded_model



import pathlib
from setuptools import setup, find_packages
import os

# The directory containing this file
HERE = pathlib.Path(__file__).parent
# The text of the README file
with open(os.path.join(HERE, "README.md"), "r") as f:
    README = f.read()

# This call to setup() does all the work
setup(
    name="enda",
    version="0.0.2",

    description="Tools to manipulate energy time-series and contracts, and to perform forecasts.",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/enercoop/enda",
    author="Enercoop",
    author_email="emmanuel.charon@enercoop.org",
    license="MIT",
    classifiers=[
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ],
    project_urls={
        "Bug Tracker": "https://github.com/enercoop/enda/issues"
    },

    packages=find_packages(exclude=("tests",)),
    include_package_data=True,
    python_requires='>=3.7.3',
    install_requires=[
        "pandas>=1.0.0"
    ]
)



from dateutil.relativedelta import relativedelta


class TimezoneUtils:

    @staticmethod
    def is_timezone_aware(dt):
        return dt.tzinfo is not None and dt.tzinfo.utcoffset(dt) is not None

    @classmethod
    def add_interval_to_day_dt(cls, day_dt, interval):
        """ Adds an interval (not more precise than a day) to a day,
        correctly dealing with timezone-aware (and naive) day_dt;
        works around daylight savings time changes.

       Normally, to add an interval to a day which is not timezone aware, simply use:
           day_dt + interval.
       This does not work properly for timezone-aware days so we added this function.

       :param day_dt: a timezone_aware datetime which is a day (hour=minute=seconds=microsecond=0)
       :param interval: a interval of type relativedelta not more precise than a day
       """

        if not (day_dt.hour == day_dt.minute == day_dt.second == day_dt.microsecond == 0):
            raise ValueError("day_dt must be datetime with only years, months or days (not more precise),"
                             " but given: {}, {}".format(type(day_dt), day_dt))
        if not (isinstance(interval, relativedelta)
                and interval.hours == interval.minutes == interval.seconds == interval.microseconds == 0):
            raise(ValueError("Interval must be a relativedelta with only years, months or days "
                             "(not more precise), but given: {}, {}".format(type(interval), interval)))

        tz = day_dt.tzinfo
        day_naive = day_dt.replace(tzinfo=None)
        day_naive = day_naive + interval

        if tz is not None:
            day_aware = tz.localize(day_naive, is_dst=None)
            return day_aware
        else:
            return day_naive
